@@ -2,7 +2,7 @@ import json
 import io
 import os
 from functools import wraps
-from typing import Any, Dict, Generator, List, Optional, Union
+from typing import Any, Dict, List, Optional
 
 import duckdb
 import pandas as pd
@@ -100,7 +100,7 @@ class HdDB:
                     tm.id, 
                     tm.label, 
                     ic.table_name AS tbl, 
-                    ic.data_type AS type
+                    'Txt' AS type
                 FROM 
                     temp_metadata tm
                 JOIN 
@@ -152,7 +152,7 @@ class HdDB:
             logger.info(f"Database {org}__{db} successfully deleted from Motherduck")
         except duckdb.Error as e:
             logger.error(f"Error deleting database from MotherDuck: {e}")
-            raise ConnectionError(f"Error deleting database from MotherDuck: {e}")
+            raise QueryError(f"Error deleting database from MotherDuck: {e}")
 
     @attach_motherduck
     def get_data(self, org: str, db: str, tbl: str) -> dict:
@@ -181,7 +181,7 @@ class HdDB:
             return {"data": data_json, "fields": fields_json}
         except duckdb.Error as e:
             logger.error(f"Error retrieving data from MotherDuck: {e}")
-            raise ConnectionError(f"Error retrieving data from MotherDuck: {e}")
+            raise QueryError(f"Error retrieving data from MotherDuck: {e}")
 
     @attach_motherduck
     def drop_table(self, org: str, db: str, tbl: str):
@@ -216,7 +216,7 @@ class HdDB:
             )
         except duckdb.Error as e:
             logger.error(f"Error deleting table from MotherDuck: {e}")
-            raise ConnectionError(f"Error deleting table from MotherDuck: {e}")
+            raise QueryError(f"Error deleting table from MotherDuck: {e}")
 
     @attach_motherduck
     def add_table(self, org: str, db: str, tbl: str, df: pd.DataFrame):
@@ -302,7 +302,7 @@ class HdDB:
         except duckdb.Error as e:
             self.execute("ROLLBACK;")
             logger.error(f"Error adding table to MotherDuck: {e}")
-            raise ConnectionError(f"Error adding table to MotherDuck: {e}")
+            raise QueryError(f"Error adding table to MotherDuck: {e}")
 
     @attach_motherduck
     def download_data(
@@ -363,7 +363,7 @@ class HdDB:
             return buffer
 
         except duckdb.Error as e:
-            logger.error(f"Error downloading/exporting data from table {tbl}: {e}")
+            logger.error(f"Error downloading / exporting data from table {tbl}: {e}")
             raise
 
     @attach_motherduck
@@ -388,7 +388,7 @@ class HdDB:
             return True
         except duckdb.Error as e:
             logger.error(f"Error updating data in MotherDuck: {e}")
-            raise ConnectionError(f"Error updating data in MotherDuck: {e}")
+            raise QueryError(f"Error updating data in MotherDuck: {e}")
 
     def close(self):
         try:
@@ -416,9 +416,93 @@ class HdDB:
             query = f'UPDATE "{org}__{db}".hd_tables SET nrow = nrow - 1 WHERE id = ?'
             self.execute(query, [tbl])
             self.execute("COMMIT;")
-            logger.info(f"Row with rcd___id {rcd___id} successfully deleted from table {tbl}")
+            logger.info(
+                f"Row with rcd___id {rcd___id} successfully deleted from table {tbl}"
+            )
             return True
         except duckdb.Error as e:
             self.execute("ROLLBACK;")
             logger.error(f"Error deleting data in MotherDuck: {e}")
-            raise ConnectionError(f"Error deleting data in MotherDuck: {e}")
+            raise QueryError(f"Error deleting data in MotherDuck: {e}")
+
+    @attach_motherduck
+    def add_row(self, org: str, db: str, tbl: str, row: dict):
+        try:
+            columns = ", ".join(f'"{k}"' for k in row.keys())
+            placeholders = ", ".join(["?" for _ in row])
+
+            self.execute("BEGIN TRANSACTION;")
+            query = (
+                f'INSERT INTO "{org}__{db}"."{tbl}" ({columns}) VALUES ({placeholders})'
+            )
+            self.execute(query, list(row.values()))
+            self.execute(
+                f'UPDATE "{org}__{db}".hd_tables SET nrow = nrow + 1 WHERE id = ?',
+                [tbl],
+            )
+            self.execute("COMMIT;")
+            return True
+        except duckdb.Error as e:
+            self.execute("ROLLBACK;")
+            logger.error(f"Error adding row to table {tbl}: {e}")
+            raise QueryError(f"Error adding row to table {tbl}: {e}")
+
+    @attach_motherduck
+    def add_column(self, org: str, db: str, tbl: str, column: dict):
+        try:
+            self.execute("BEGIN TRANSACTION;")
+            self.execute(
+                f'ALTER TABLE "{org}__{db}"."{tbl}" ADD COLUMN "{column["slug"]}" VARCHAR'
+            )
+            self.execute(
+                f'UPDATE "{org}__{db}".hd_tables SET ncol = ncol + 1 WHERE id = ?',
+                [tbl],
+            )
+            self.execute(
+                f'INSERT INTO "{org}__{db}".hd_fields (fld___id, id, label, tbl, type) VALUES (?, ?, ?, ?, ?)',
+                [
+                    column["fld___id"],
+                    column["slug"],
+                    column["headerName"],
+                    tbl,
+                    column["type"],
+                ],
+            )
+            self.execute("COMMIT;")
+            return True
+        except duckdb.Error as e:
+            self.execute("ROLLBACK;")
+            logger.error(f"Error adding column to table {tbl}: {e}")
+            raise QueryError(f"Error adding column to table {tbl}: {e}")
+        
+    @attach_motherduck
+    def delete_column(self, org: str, db: str, tbl: str, column: dict):
+        try:
+            column_name = column["slug"]
+            column_id = column["fld___id"]
+            self.execute("BEGIN TRANSACTION;")
+            self.execute(f'ALTER TABLE "{org}__{db}"."{tbl}" DROP COLUMN "{column_name}"')
+            self.execute(
+                f'UPDATE "{org}__{db}".hd_tables SET ncol = ncol - 1 WHERE id = ?',
+                [tbl],
+            )
+            self.execute(
+                f'DELETE FROM "{org}__{db}".hd_fields WHERE id = ? AND tbl = ?',
+                [column_id, tbl],
+            )
+            self.execute("COMMIT;")
+            return True
+        except duckdb.Error as e:
+            self.execute("ROLLBACK;")
+            logger.error(f"Error deleting column from table {tbl}: {e}")
+            raise QueryError(f"Error deleting column from table {tbl}: {e}")
+    
+
+    @attach_motherduck
+    def update_hd_fields(self, org: str, db: str, fld___id: str, label: str, type: str):
+        try:
+            query = f'UPDATE "{org}__{db}".hd_fields SET label = ?, type = ? WHERE fld___id = ?'
+            self.execute(query, [label, type, fld___id])
+        except duckdb.Error as e:
+            logger.error(f"Error updating hd_fields: {e}")
+            raise QueryError(f"Error updating hd_fields: {e}")
